import logging
import zipfile
from io import BytesIO
from typing import Union, Optional, List

import cv2
import numpy as np
import robothub_core

__all__ = ['send_image_event', 'send_frame_event_with_zipped_images']

logger = logging.getLogger(__name__)


def _log_event_status(result: bool, event_id):
    if result:
        logger.info(f"Event {event_id}: sent successfully.")
    else:
        logger.info(f'Event {event_id}: failed to send.')


def send_image_event(image: Union[np.ndarray, bytes],
                     title: str,
                     device_id: str = None,
                     metadata: Optional[dict] = None,
                     tags: List[str] = None,
                     mjpeg_quality=98,
                     encode=False
                     ) -> Optional[str]:
    """Send a single image frame event to RH."""
    try:
        if encode:
            _, image = cv2.imencode(".jpg", image, [int(cv2.IMWRITE_JPEG_QUALITY), mjpeg_quality])

        event = robothub_core.EVENTS.prepare()
        event.add_frame(bytes(image), device_id)
        event.set_title(title)
        if metadata:
            event.set_metadata(metadata)
        if tags:
            event.add_tags(tags)
        robothub_core.EVENTS.upload(event)
        _log_event_status(True, event.id)
        return event.id
    except Exception as e:
        logger.error(f'Failed to send event: {e}')
        _log_event_status(False, "None")
        return None


def send_frame_event_with_zipped_images(cv_frame,
                                        files,
                                        title: str,
                                        device_id: str,
                                        tags: List[str] = None,
                                        metadata: Optional[dict] = None,
                                        encode: bool = False,
                                        mjpeg_quality=98
                                        ) -> Optional[str]:
    """Send a collection of images as a single event to RH."""
    try:
        encode_param = [int(cv2.IMWRITE_JPEG_QUALITY), mjpeg_quality]
        if encode:
            _, cv_frame = cv2.imencode('.jpg', cv_frame, encode_param)

        event = robothub_core.EVENTS.prepare()
        event.add_frame(bytes(cv_frame), device_id)
        event.set_title(title)
        if tags:
            event.set_tags(tags)
        if metadata:
            event.set_metadata(metadata)
        logger.debug(f'Total files: {len(files)}')
        with BytesIO() as zip_buffer:
            with zipfile.ZipFile(zip_buffer, 'w') as zip_file:
                for idx, file in enumerate(files):
                    if encode:
                        _, encoded = cv2.imencode('.jpg', file, encode_param)
                    else:
                        encoded = file
                    # Convert the numpy array to bytes
                    image_bytes = encoded.tobytes()
                    # Add the bytes to the zip file with a unique filename
                    filename = f'image_{idx}.jpeg'
                    zip_file.writestr(filename, image_bytes)
                    # event.add_file(bytes(encoded), name=f"image_{idx}.jpeg")
            # Get the bytes of the zip file
            zip_bytes = zip_buffer.getvalue()
            event.add_file(zip_bytes, name=f'images.zip')
            robothub_core.EVENTS.upload(event)
            _log_event_status(True, event.id)
            return event.id
    except Exception as e:
        logger.error(f'Failed to send event: {e}')
        _log_event_status(False, 'None')
        return None



import time
from typing import List, Optional, Union, Dict, Tuple

import depthai as dai
import numpy as np
import robothub_core
from depthai_sdk import OakCamera
from depthai_sdk.components import Component, CameraComponent, StereoComponent, NNComponent
from depthai_sdk.oak_outputs.xout.xout_base import StreamXout
from depthai_sdk.oak_outputs.xout.xout_h26x import XoutH26x
from depthai_sdk.visualize.objects import VisText, VisLine

from robothub_oak.types import BoundingBox

__all__ = ['LiveView', 'LIVE_VIEWS']


def _create_stream_handle(camera_serial: str, unique_key: str, name: str):
    if unique_key not in robothub_core.STREAMS.streams:
        color_stream_handle = robothub_core.STREAMS.create_video(camera_serial, unique_key, name)
    else:
        color_stream_handle = robothub_core.STREAMS.streams[unique_key]
    return color_stream_handle


def _publish_data(stream_handle: robothub_core.StreamHandle,
                  h264_frame,
                  rectangles: List[BoundingBox],
                  rectangle_labels: List[str],
                  texts: List[VisText],
                  lines: List[VisLine],
                  frame_width: int,
                  frame_height: int):
    timestamp = int(time.perf_counter_ns() / 1e6)
    metadata = {
        "platform": "robothub",
        "frame_shape": [frame_height, frame_width],
        "config": {
            "output": {
                "img_scale": 1.0,
                "show_fps": False,
                "clickable": True
            },
            "detection": {
                "thickness": 1,
                "fill_transparency": 0.05,
                "box_roundness": 0,
                "color": [0, 255, 0],
                "bbox_style": 0,
                "line_width": 0.5,
                "line_height": 0.5,
                "hide_label": False,
                "label_position": 0,
                "label_padding": 10
            },
            'text': {
                'font_color': [255, 255, 0],
                'font_transparency': 0.5,
                'font_scale': 1.0,
                'font_thickness': 2,
                'bg_transparency': 0.5,
                'bg_color': [0, 0, 0]
            }
        },
        "objects": [
            {
                "type": "detections",
                "detections": []
            }
        ]
    }

    # Bounding boxes
    for roi, label in zip(rectangles, rectangle_labels):
        xmin, ymin, xmax, ymax = roi
        metadata['objects'][0]['detections'].append(
            {'bbox': [xmin, ymin, xmax, ymax], 'label': label, 'color': [0, 255, 255]}
        )

    # Texts
    for text in texts:
        metadata["objects"].append(text.prepare().serialize())

    # Lines
    for line in lines:
        metadata["objects"].append(line.prepare().serialize())

    # Publish
    stream_handle.publish_video_data(bytes(h264_frame), timestamp, metadata)


class LiveView:
    def __init__(self,
                 name: str,
                 unique_key: str,
                 device_mxid: str,
                 frame_width: int,
                 frame_height: int):
        """
        :param name: Name of the Live View.
        :param unique_key: Live View identifier.
        :param device_mxid: MXID of the device that is streaming the Live View.
        :param frame_width: Frame width.
        :param frame_height: Frame height.
        """
        self.frame_width = frame_width
        self.frame_height = frame_height
        self.device_mxid = device_mxid
        self.unique_key = unique_key
        self.name = name

        self.stream_handle = _create_stream_handle(camera_serial=device_mxid, unique_key=unique_key, name=name)

        # Objects
        self.texts: List[VisText] = []
        self.rectangles: List[BoundingBox] = []
        self.labels: List[str] = []
        self.lines: List[VisLine] = []

    @staticmethod
    def create(device: OakCamera,
               component: Union[CameraComponent, StereoComponent],
               name: str,
               unique_key: str = None,
               manual_publish: bool = False
               ) -> 'LiveView':
        """
        Creates a Live View for a given component.

        :param device: OakCamera instance.
        :param component: Component to create a Live View for. Either a CameraComponent, StereoComponent or NNComponent.
        :param name: Name of the Live View.
        :param unique_key: Live View identifier.
        :param manual_publish: If True, the Live View will not be automatically published. Use LiveView.publish() to publish the Live View.
        """
        output = None
        is_h264 = LiveView.is_encoder_enabled(component)
        if not is_h264:
            output = LiveView.h264_output(device, component)
        elif not is_h264 and not isinstance(component, CameraComponent):
            raise ValueError(f'Component {component.__class__.__name__} must have h264 encoding '
                             f'enabled to be used with LiveView.')

        w, h = LiveView.get_stream_size(component)
        device_mxid = device.device.getMxId()
        unique_key = unique_key or f'{device_mxid}_{component.__class__.__name__.lower()}_encoded'

        live_view = LiveView(name=name,
                             unique_key=unique_key,
                             device_mxid=device_mxid,
                             frame_width=w,
                             frame_height=h)

        if not manual_publish:
            device.callback(output or component.out.encoded, live_view.h264_callback)

        LIVE_VIEWS[unique_key] = live_view
        return live_view

    @staticmethod
    def h264_output(device: OakCamera, component: CameraComponent):
        fps = 30
        if isinstance(component, StereoComponent):
            fps = component._fps
        elif isinstance(component, CameraComponent):
            fps = component.get_fps()

        encoder = device.pipeline.createVideoEncoder()
        encoder_profile = dai.VideoEncoderProperties.Profile.H264_MAIN
        encoder.setDefaultProfilePreset(fps, encoder_profile)
        encoder.input.setQueueSize(1)
        encoder.input.setBlocking(False)
        encoder.setKeyframeFrequency(int(fps))
        encoder.setBitrate(1500 * 1000)
        encoder.setRateControlMode(dai.VideoEncoderProperties.RateControlMode.CBR)
        encoder.setNumFramesPool(3)

        component.node.video.link(encoder.input)

        def encoded(pipeline, device):
            xout = XoutH26x(
                frames=StreamXout(encoder.id, encoder.bitstream),
                color=True,
                profile=encoder_profile,
                fps=encoder.getFrameRate(),
                frame_shape=component.node.getResolution()
            )
            xout.name = component._source
            return component._create_xout(pipeline, xout)

        return encoded

    @staticmethod
    def is_encoder_enabled(component: Component) -> bool:
        """
        Checks if the component has h264 encoding enabled.

        :param component: Component to check.
        :return: True if the component has h264 encoding enabled, False otherwise.
        :raises ValueError: If the component is not a CameraComponent or StereoComponent.
        """
        if not isinstance(component, CameraComponent) and not isinstance(component, StereoComponent):
            raise ValueError(f'Component {component.__class__.__name__} must be a CameraComponent or StereoComponent.')
        encoder = component.encoder

        if encoder and encoder.getProfile() != dai.VideoEncoderProperties.Profile.H264_MAIN:
            return False

        return True

    @staticmethod
    def get_stream_size(component) -> Tuple[int, int]:
        if isinstance(component, CameraComponent):
            return component.stream_size
        elif isinstance(component, StereoComponent):
            return component.left.stream_size
        elif isinstance(component, NNComponent):
            return component._input.stream_size

    @staticmethod
    def get(unique_key: str = None, name: str = None):
        if name is not None:
            return LiveView.get_by_name(name)
        elif unique_key is not None:
            return LiveView.get_by_unique_key(unique_key)
        else:
            raise ValueError('Either name or unique_key must be specified.')

    @staticmethod
    def get_by_name(name: str) -> Optional['LiveView']:
        for live_view in LIVE_VIEWS.values():
            if live_view.name == name:
                return live_view
        return None

    @staticmethod
    def get_by_unique_key(unique_key: str) -> Optional['LiveView']:
        if unique_key not in LIVE_VIEWS:
            raise ValueError(f'Live View with unique_key {unique_key} does not exist.')

        return LIVE_VIEWS[unique_key]

    def add_rectangle(self, rectangle: BoundingBox, label: str) -> None:
        self.rectangles.append(rectangle)
        self.labels.append(label)

    def add_text(self,
                 text: str,
                 coords: Tuple[int, int],
                 size: int = None,
                 color: Tuple[int, int, int] = None,
                 thickness: int = None,
                 outline: bool = True,
                 background_color: Tuple[int, int, int] = None,
                 background_transparency: float = 0.5) -> None:
        obj = VisText(text, coords, size, color, thickness, outline, background_color, background_transparency)
        self.texts.append(obj)

    def add_line(self,
                 pt1: Tuple[int, int],
                 pt2: Tuple[int, int],
                 color: Tuple[int, int, int] = None,
                 thickness: int = None
                 ) -> None:
        obj = VisLine(pt1, pt2, color, thickness)
        self.lines.append(obj)

    def publish(self, h264_frame: Union[np.array, List]) -> None:
        _publish_data(stream_handle=self.stream_handle,
                      h264_frame=h264_frame,
                      rectangles=self.rectangles,
                      rectangle_labels=self.labels,
                      texts=self.texts,
                      lines=self.lines,
                      frame_width=self.frame_width,
                      frame_height=self.frame_height)
        self._reset_overlays()

    def h264_callback(self, h264_packet):
        self.publish(h264_frame=h264_packet.frame)

    def _reset_overlays(self):
        self.rectangles.clear()
        self.labels.clear()
        self.lines.clear()
        self.texts.clear()


LIVE_VIEWS: Dict[str, LiveView] = dict()



import logging
from typing import Dict, Any

import depthai
import robothub_core

__all__ = ['setup_logger', 'get_device_performance_metrics', 'get_device_details', 'try_or_default']


def setup_logger(name: str, level: int = logging.INFO):
    """
    :param name: Logger name.
    :param level: Either a string or an integer. If a string, it must be one of the following: 'DEBUG', 'INFO',
    'WARNING', 'ERROR', 'CRITICAL'. If an integer, it must be one of the following: log.DEBUG, log.INFO, log.WARNING,
    log.ERROR, log.CRITICAL.
    """
    if isinstance(level, str):
        level = level.upper()
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.propagate = False

    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    format_str = '%(levelname)s | %(name)s | %(message)s' if level == logging.DEBUG else '%(levelname)s | %(message)s'
    formatter = logging.Formatter(format_str)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def get_device_performance_metrics(device: depthai.Device) -> Dict[str, Any]:
    """
    Returns a dictionary with statistics about the device.
    """
    stats = {'mxid': device.getMxId()}

    css_cpu_usage = device.getLeonCssCpuUsage().average
    mss_cpu_usage = device.getLeonMssCpuUsage().average
    cmx_mem_usage = device.getCmxMemoryUsage()
    ddr_mem_usage = device.getDdrMemoryUsage()
    chip_temp = device.getChipTemperature()

    stats['css_usage'] = int(100 * css_cpu_usage)
    stats['mss_usage'] = int(100 * mss_cpu_usage)
    stats['ddr_mem_free'] = int(ddr_mem_usage.total - ddr_mem_usage.used)
    stats['ddr_mem_total'] = int(ddr_mem_usage.total)
    stats['cmx_mem_free'] = int(cmx_mem_usage.total - cmx_mem_usage.used)
    stats['cmx_mem_total'] = int(cmx_mem_usage.total)
    stats['css_temp'] = int(100 * chip_temp.css)
    stats['mss_temp'] = int(100 * chip_temp.mss)
    stats['upa_temp'] = int(100 * chip_temp.upa)
    stats['dss_temp'] = int(100 * chip_temp.dss)
    stats['temp'] = int(100 * chip_temp.average)

    return stats


def get_device_details(device: depthai.Device, state: robothub_core.DeviceState) -> Dict[str, Any]:
    """
    Returns a dictionary with information about the device.
    """
    info = {
        'mxid': device.getMxId(),
        'protocol': 'unknown',
        'platform': 'unknown',
        'product_name': 'unknown',
        'board_name': 'unknown',
        'board_rev': 'unknown',
        'bootloader_version': 'unknown',
        'usb_speed': 'unknown',
        'state': state.value
    }

    device_info = try_or_default(device.getDeviceInfo)
    calibration = try_or_default(device.readFactoryCalibration) or try_or_default(device.readCalibration2)
    eeprom_data = try_or_default(calibration.getEepromData)

    info['usb_speed'] = device.getUsbSpeed().name.lower()
    bootloader_version = device.getBootloaderVersion()  # can be None
    if bootloader_version:
        info['bootloader_version'] = bootloader_version.toStringSemver()

    if eeprom_data:
        info['product_name'] = eeprom_data.productName
        info['board_name'] = eeprom_data.boardName
        info['board_rev'] = eeprom_data.boardRev

    if device_info:
        info['protocol'] = device_info.protocol.name
        info['platform'] = device_info.platform.name

    return info


def try_or_default(func, default=None):
    try:
        return func()
    except:
        return default



import atexit
import logging
import signal
import threading
import time

from abc import ABC, abstractmethod
from threading import Thread
from typing import Optional

import robothub_core
from depthai_sdk import OakCamera

from robothub_oak.utils import get_device_performance_metrics, get_device_details

__all__ = ["BaseApplication"]

logger = logging.getLogger(__name__)


class BaseApplication(robothub_core.RobotHubApplication, ABC):
    """
    This class acts as the main entry point for the user, managing a single device, creating pipelines,
    and polling the device for new data. Derived classes must implement the `setup_pipeline` method.

    Attributes:
        config: The configuration settings from the robotapp.toml config file.

    Methods:
        setup_pipeline: Abstract method to be implemented by child classes. Sets up the pipeline for a device.
        on_device_connected: Optional method, called when a device is connected.
        on_device_disconnected: Optional method, called when a device is disconnected.
        on_stop: Optional method, called when the application is stopped.
    """

    def __init__(self):
        robothub_core.RobotHubApplication.__init__(self)
        ABC.__init__(self)

        self.config = robothub_core.CONFIGURATION

        self.__rh_device: Optional[robothub_core.RobotHubDevice] = None
        self.__device_mxid: Optional[str] = None
        self.__device_product_name: Optional[str] = None
        self.__device_state: Optional[robothub_core.DeviceState] = None
        self.__device: Optional[OakCamera] = None
        self.__device_thread: Optional[Thread] = None

        atexit.register(self.__cleanup)
        signal.signal(signal.SIGUSR1, self.__cleanup)

        self.__manage_event = threading.Event()
        self.__report_event = threading.Event()

    def on_start(self) -> None:
        if len(robothub_core.DEVICES) == 0:
            logger.info("No assigned devices.")
            self.stop_event.set()
            return
        if len(robothub_core.DEVICES) > 1:
            logger.warning("More than one device assigned, only the first one will be used.")

        self.__rh_device = robothub_core.DEVICES[0]
        self.__device_mxid = self.__rh_device.oak["serialNumber"]
        self.__device_product_name = (
                self.__rh_device.oak.get("name", None)
                or self.__rh_device.oak.get("productName", None)
                or self.__device_mxid
        )
        self.__device_state = robothub_core.DeviceState.DISCONNECTED

        self.__device_thread = Thread(
            target=self.__manage_device,
            name=f"manage_{self.__device_mxid}",
            daemon=False,
        )
        self.__device_thread.start()

    def start_execution(self) -> None:
        self.stop_event.wait()  # Keep main thread alive

    @abstractmethod
    def setup_pipeline(self, oak: OakCamera) -> None:
        """
        The entry point for the application. This method is called when a device is connected and ready to be used.

        :param oak: The device that is ready to be used.
        """
        pass

    def on_device_connected(self, oak: OakCamera) -> None:
        """
        Called when a camera is connected.

        :param oak: The camera that was connected.
        """
        pass

    def on_device_disconnected(self) -> None:
        """
        Called when a camera is disconnected. Opposite of on_device_connected.
        """
        pass

    def __cleanup(self, *args, **kwargs) -> None:
        """
        Called when the application is stopped. Registered as atexit handler.
        """
        # Device thread must close the device
        self.stop_event.set()
        self.__manage_event.set()
        self.__report_event.set()

        self.__device_thread.join()

    def __manage_device(self) -> None:
        """
        Handle the life cycle of the device.
        """
        logger.debug(f"Device {self.__device_product_name}: management thread started.")

        while self.running:
            # if device is not connected
            if self.__device is None or not self.__device.running():
                # Make sure it is properly closed in case it disconnected during runtime
                self.__close_device()
                self.on_device_disconnected()

                # Connect to the device
                self.__connect()

                # If device is connected
                if self.__device:
                    logger.debug(f"Device {self.__device_product_name}: creating pipeline...")

                    self.setup_pipeline(oak=self.__device)
                    self.__device.start(blocking=False)
                    self.on_device_connected(self.__device)

                    logger.info(f"Device {self.__device_product_name}: started successfully.")

                    self.__manage_event.clear()
                    self.__report_event.clear()

                    # Threads for polling and reporting
                    polling_thread = Thread(
                        target=self.__poll_device,
                        daemon=True,
                        name=f"poll_{self.__device_mxid}",
                    )
                    polling_thread.start()

                    reporting_thread = Thread(
                        target=self.__device_stats_reporting,
                        daemon=False,
                        name=f"stats_reporting_{self.__device_mxid}",
                    )
                    reporting_thread.start()
                else:
                    self.__manage_event.wait(25)

            self.__manage_event.wait(5)

        # Make sure device is closed
        self.__close_device()
        self.on_device_disconnected()
        logger.debug(f"Device {self.__device_product_name}: thread stopped.")

    def __poll_device(self) -> None:
        """
        Poll the device for new data. This method is called in a separate thread.
        """
        while self.running and not self.__manage_event.is_set():
            if not self.__device.poll():
                return

            time.sleep(0.0025)

    def __device_stats_reporting(self) -> None:
        """
        Report device info and stats every 30 seconds.
        """
        dai_device = self.__device.device
        state = self.__device_state
        while self.running and not self.__report_event.is_set():
            if self.__device is None or not self.__device.running():
                return

            try:
                device_info = get_device_details(dai_device, state)
                device_stats = get_device_performance_metrics(dai_device)

                robothub_core.AGENT.publish_device_info(device_info)
                robothub_core.AGENT.publish_device_stats(device_stats)
            except Exception as e:
                product_name = self.__device_product_name
                logger.debug(f"Device {product_name}: could not report info/stats with error: {e}.")

            self.__report_event.wait(30)

    def __connect(self) -> None:
        """
        Connect to the device. This method is called in a separate thread.
        """
        give_up_time = time.time() + 30

        self.__device_state = robothub_core.DeviceState.CONNECTING
        product_name = self.__device_product_name
        while time.time() < give_up_time and self.running:
            logger.debug(
                f"Device {product_name}: remaining time to connect - {give_up_time - time.time()} seconds."
            )
            try:
                oak = OakCamera(self.__device_mxid)
                self.__device = oak
                self.__device_state = robothub_core.DeviceState.CONNECTED
                logger.debug(f"Device {product_name}: successfully connected.")
                return
            except Exception as e:
                # If device can't be connected to on first try, wait 5 seconds and try again.
                logger.debug(f"Device {product_name}: error while trying to connect - {e}.")
                self.wait(5)

        logger.info(f"Device {product_name}: could not manage to connect within 30s timeout.")
        self.__device = None
        self.__device_state = robothub_core.DeviceState.DISCONNECTED
        return

    def __close_device(self):
        """
        Close the device gracefully. If the device is not running, this method does nothing.
        """
        if self.__device is None or not self.__device.running():
            return

        self.__device.__exit__(1, 2, 3)
        self.__device = None
        product_name = self.__device_product_name
        logger.info(f"Device {product_name}: closed gracefully.")

    def get_device(self) -> Optional[OakCamera]:
        """
        Get a device by its mxid. If the device is not running, this method returns None.
        :return: The device or None if the device is not running.
        """
        return self.__device

    def restart_device(self):
        """
        Restart the device.
        """
        if not self.__device:
            logger.warning(f"Device is not initialized and cannot be restarted.")
            return

        self.__close_device()
        self.__report_event.set()
        self.__manage_event.set()



import io

from setuptools import setup

long_description = io.open('README.md', encoding='utf-8').read()

setup(
    name='robothub_oak',
    version='2.1.0',
    description='RobotHub integration library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://www.luxonis.com/',
    license='MIT',
    keywords='robothub camera robot hub connect agent depthai sdk',
    author='Luxonis',
    author_email='support@luxonis.com',
    packages=['robothub_oak'],
    package_dir={'': 'src'},  # https://stackoverflow.com/a/67238346/5494277
    include_package_data=True,
    project_urls={
        'Homepage': 'https://github.com/luxonis/robothub/',
        'Documentation': 'https://hub-docs.luxonis.com/',
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
    ],
)



import os

from robothub_oak.application import *
from robothub_oak.live_view import *
from robothub_oak.utils import setup_logger

try:
    import blobconverter

    # Suppress blobconverter logs
    blobconverter.set_defaults(silent=True)
except ImportError:
    pass

__version__ = '2.1.0'

REPLAY_PATH = os.environ.get('RH_OAK_REPLAY_PATH', None) or os.environ.get('RH_REPLAY_PATH', None)

# Setup logging for the module
setup_logger(__name__)
